﻿/**
* ******************************************************
*                                                       *
*   Copyright (C) Microsoft. All rights reserved.       *
*                                                       *
*******************************************************
*/
/// <reference path="../../../typings/node.d.ts" />
/// <reference path="../../../typings/Q.d.ts" />
/// <reference path="../../../typings/node.d.ts" />
/// <reference path="../../../typings/rimraf.d.ts" />
"use strict";
var child_process = require("child_process");
var fs = require("fs");
var nconf = require("nconf");
var os = require("os");
var path = require("path");
var Q = require("q");
var readline = require("readline");
var rimraf = require("rimraf");

var utils = require("vcremote-utils");

var resources = utils.ResourcesManager;
var utilHelper = utils.UtilHelper;

var exec = child_process.exec;

var Certs;
(function (Certs) {
    var debug = false;
    var CERT_DEFAULTS;
    (function (CERT_DEFAULTS) {
        CERT_DEFAULTS.days = 1825;
        CERT_DEFAULTS.country = "US";
        CERT_DEFAULTS.ca_cn = "vcremote-Certificate-Authority";
        CERT_DEFAULTS.pfx_name = "vcremote-Client-Certificate";
    })(CERT_DEFAULTS || (CERT_DEFAULTS = {}));
    ;

    function resetServerCert(conf, yesOrNoHandler) {
        var certsDir = path.join(conf.get("serverDir"), "certs");

        if (!fs.existsSync(certsDir)) {
            return initializeServerCerts(conf);
        }

        var shouldProceedDeferred = Q.defer();
        yesOrNoHandler = yesOrNoHandler || readline.createInterface({ input: process.stdin, output: process.stdout });
        var answerCallback = function (answer) {
            answer = answer.toLowerCase();
            if (answer === "y" || answer === "yes") {
                yesOrNoHandler.close();
                shouldProceedDeferred.resolve(true);
            } else if (answer === "n" || answer === "no") {
                yesOrNoHandler.close();
                shouldProceedDeferred.resolve(false);
            } else {
                yesOrNoHandler.question("Please answer [Y]es or [N]o." + os.EOL, answerCallback);
            }
        };
        yesOrNoHandler.question("Warning: This command deletes all current client certificates. " + "You will need to generate and configure a new security PIN. Do you want to continue? [Y]es or [N]o" + os.EOL, answerCallback);
        return shouldProceedDeferred.promise.then(function (shouldProceed) {
            if (shouldProceed) {
                rimraf.sync(certsDir);
                return initializeServerCerts(conf);
            }

            return Q({});
        });
    }
    Certs.resetServerCert = resetServerCert;
    ;

    function generateClientCert(conf) {
        var certsDir = path.join(conf.get("serverDir"), "certs");
        var caKeyPath = path.join(certsDir, "ca-key.pem");
        var caCertPath = path.join(certsDir, "ca-cert.pem");
        if (!fs.existsSync(caKeyPath) || !fs.existsSync(caCertPath)) {
            var error = resources.getStringForLanguage(conf.get("lang"), "CAFilesNotFound", caKeyPath, caCertPath);
            return Q(0).thenReject(error);
        }

        return makeClientPinAndSslCert(caKeyPath, caCertPath, certsDir, certOptionsFromConf(conf), conf).then(function (pin) {
            if (utilHelper.argToBool(conf.get("suppressVisualStudioMessage"))) {
                return pin;
            }

            printVisualStudioSetupToConsole(conf, pin);
            return pin;
        });
    }
    Certs.generateClientCert = generateClientCert;
    ;

    function initializeServerCerts(conf) {
        var certsDir = path.join(conf.get("serverDir"), "certs");
        var certPaths = {
            certsDir: certsDir,
            caKeyPath: path.join(certsDir, "ca-key.pem"),
            caCertPath: path.join(certsDir, "ca-cert.pem"),
            serverKeyPath: path.join(certsDir, "server-key.pem"),
            serverCertPath: path.join(certsDir, "server-cert.pem"),
            newCerts: false
        };

        var certsExist = fs.existsSync(certPaths.caCertPath) && fs.existsSync(certPaths.serverKeyPath) && fs.existsSync(certPaths.serverCertPath);
        certPaths.newCerts = !certsExist;
        var promise;
        if (certsExist) {
            promise = isExpired(certPaths.caCertPath).then(function (hasExpired) {
                if (hasExpired) {
                    return Q(true);
                }

                return isExpired(certPaths.serverCertPath);
            });
        } else {
            promise = Q(true); // do not exist, so true -> need making
        }

        promise = promise.then(function (shouldMake) {
            if (!shouldMake) {
                return Q({});
            }

            utilHelper.createDirectoryIfNecessary(certsDir);
            var options = certOptionsFromConf(conf);
            return makeSelfSigningCACert(certPaths.caKeyPath, certPaths.caCertPath, options).then(function () {
                return makeSelfSignedCert(certPaths.caKeyPath, certPaths.caCertPath, certPaths.serverKeyPath, certPaths.serverCertPath, options, conf);
            }).then(function () {
                certPaths.newCerts = true;
            });
        }).then(function () {
            return {
                newCerts: certPaths.newCerts,
                getKey: function () {
                    return fs.readFileSync(certPaths.serverKeyPath);
                },
                getCert: function () {
                    return fs.readFileSync(certPaths.serverCertPath);
                },
                getCA: function () {
                    return fs.readFileSync(certPaths.caCertPath);
                }
            };
        });
        return promise;
    }
    Certs.initializeServerCerts = initializeServerCerts;

    function getServerCerts(conf) {
        var certsDir = path.join(conf.get("serverDir"), "certs");
        var certPaths = {
            caCertPath: path.join(certsDir, "ca-cert.pem"),
            serverKeyPath: path.join(certsDir, "server-key.pem"),
            serverCertPath: path.join(certsDir, "server-cert.pem")
        };
        var certsExist = fs.existsSync(certPaths.caCertPath) && fs.existsSync(certPaths.serverKeyPath) && fs.existsSync(certPaths.serverCertPath);
        if (!certsExist) {
            return {
                newCerts: false,
                getKey: function () {
                    throw new Error("Not Supported");
                },
                getCert: function () {
                    throw new Error("Not Supported");
                },
                getCA: function () {
                    throw new Error("Not Supported");
                }
            };
        }
        return {
            newCerts: false,
            getKey: function () {
                return fs.readFileSync(certPaths.serverKeyPath);
            },
            getCert: function () {
                return fs.readFileSync(certPaths.serverCertPath);
            },
            getCA: function () {
                return fs.readFileSync(certPaths.caCertPath);
            }
        };
    }
    Certs.getServerCerts = getServerCerts;

    function isExpired(certPath) {
        return displayCert(certPath, ["dates"]).then(function (output) {
            var notAfter = new Date(output.stdout.substring(output.stdout.indexOf("notAfter=") + 9, output.stdout.length - 1));
            return (notAfter.getTime() < new Date().getTime());
        });
    }
    Certs.isExpired = isExpired;
    ;

    // display fields an array of any of these: 'subject', 'issuer', 'dates', etc. (see https://www.openssl.org/docs/apps/x509.html)
    function displayCert(certPath, displayFields) {
        // openssl x509 -noout -in selfsigned-cert.pem -subject -issuer -dates
        var args = "x509 -noout -in " + certPath;
        (displayFields || []).forEach(function (f) {
            args += " -" + f;
        });
        return openSslPromise(args);
    }
    Certs.displayCert = displayCert;
    ;

    function removeAllCertsSync(conf) {
        var certsDir = path.join(conf.get("serverDir"), "certs");
        if (fs.existsSync(certsDir)) {
            rimraf.sync(certsDir);
        }
    }
    Certs.removeAllCertsSync = removeAllCertsSync;

    function downloadClientCerts(pinString) {
        purgeExpiredPinBasedClientCertsSync();
        var clientCertsDir = path.join(nconf.get("serverDir"), "certs", "client");

        var pin = parseInt(pinString);
        if (isNaN(pin)) {
            throw { code: 400, id: "InvalidPin" };
        }

        var pinDir = path.join(clientCertsDir, "" + pin);
        var pfx = path.join(pinDir, "client.pfx");
        if (!fs.existsSync(pfx)) {
            throw { code: 404, id: "ClientCertNotFoundForPIN" };
        }

        return pfx;
    }
    Certs.downloadClientCerts = downloadClientCerts;

    function invalidatePIN(pinString) {
        var pinDir = path.join(nconf.get("serverDir"), "certs", "client", "" + parseInt(pinString));
        rimraf(pinDir, function () {
        });
    }
    Certs.invalidatePIN = invalidatePIN;

    function purgeExpiredPinBasedClientCertsSync() {
        var clientCertsDir = path.join(nconf.get("serverDir"), "certs", "client");
        if (!fs.existsSync(clientCertsDir)) {
            return;
        }

        var pinTimeoutInMinutes = nconf.get("pinTimeout");
        var expiredIfOlderThan = new Date().getTime() - (pinTimeoutInMinutes * 60 * 1000);
        fs.readdirSync(clientCertsDir).forEach(function (f) {
            var pfx = path.join(clientCertsDir, f, "client.pfx");
            if (fs.existsSync(pfx) && fs.statSync(pfx).mtime.getTime() < expiredIfOlderThan) {
                rimraf.sync(path.join(clientCertsDir, f));
            }
        });
    }
    Certs.purgeExpiredPinBasedClientCertsSync = purgeExpiredPinBasedClientCertsSync;
    ;

    // Makes a CA cert that will be used for self-signing our server and client certs.
    function makeSelfSigningCACert(caKeyPath, caCertPath, options) {
        options = options || {};
        var days = options.days || CERT_DEFAULTS.days;
        var country = options.country || CERT_DEFAULTS.country;
        var cn = options.cn || CERT_DEFAULTS.ca_cn;
        return openSslPromise("req -newkey rsa:4096 -x509 -days " + days + " -nodes -subj /C=" + country + "/CN=" + cn + " -keyout " + caKeyPath + " -out " + caCertPath);
    }
    Certs.makeSelfSigningCACert = makeSelfSigningCACert;
    ;

    // Makes a new private key and certificate signed with the CA.
    function makeSelfSignedCert(caKeyPath, caCertPath, outKeyPath, outCertPath, options, conf) {
        options = options || {};
        var csrPath = path.join(path.dirname(outCertPath), "CSR-" + path.basename(outCertPath));
        var days = options.days || CERT_DEFAULTS.days;
        var cn = options.cn || os.hostname();

        var cnfPath = path.join(conf.get("serverDir"), "certs", "openssl.cnf");
        writeConfigFile(cnfPath);

        return openSslPromise("genrsa -out " + outKeyPath + " 1024").then(function () {
            return openSslPromise("req -new -subj /CN=" + cn + " -key " + outKeyPath + " -out " + csrPath + " -config " + cnfPath);
        }).then(function () {
            return openSslPromise("x509 -req -days " + days + " -in " + csrPath + " -CA " + caCertPath + " -CAkey " + caKeyPath + " -extensions v3_req -extfile " + cnfPath + " -set_serial 01 -out " + outCertPath);
        }).then(function () {
            fs.unlinkSync(csrPath);
        });
    }
    Certs.makeSelfSignedCert = makeSelfSignedCert;
    ;

    function verifyCert(caCertPath, certPath) {
        return openSslPromise("verify -CAfile " + caCertPath + " " + certPath);
    }
    Certs.verifyCert = verifyCert;
    ;

    function openSslPromise(args) {
        var deferred = Q.defer();

        exec("openssl " + args, function (error, stdout, stderr) {
            if (debug) {
                console.info("exec openssl " + args);
                console.info("stdout: %s", stdout);
                console.info("stderr: %s", stderr);
            }

            if (error) {
                deferred.reject(error);
            } else {
                deferred.resolve({ stdout: stdout.toString(), stderr: stderr.toString() });
            }
        });

        return deferred.promise;
    }

    function certOptionsFromConf(conf) {
        var options = {};
        if (conf.get("certExpirationdays") < 1) {
            console.info(resources.getString("CertExpirationInvalid", conf.get("certExpirationdays"), CERT_DEFAULTS.days));
        } else {
            options.days = conf.get("certExpirationdays");
        }

        return options;
    }

    function writeConfigFile(cnfPath) {
        var net = os.networkInterfaces();
        var cnf = "[req]\ndistinguished_name = req_distinguished_name\nreq_extensions = v3_req\n[req_distinguished_name]\nC_default = US\n[ v3_req ]\nbasicConstraints = CA:FALSE\nkeyUsage = nonRepudiation, digitalSignature, keyEncipherment\nsubjectAltName = @alt_names\n[alt_names]\nDNS.1 = " + os.hostname() + "\n";
        var count = 1;

        for (var key in net) {
            for (var i = 0; i < net[key].length; i++) {
                if (net[key][i].address && !net[key][i].internal) {
                    cnf += "IP." + count + " = " + net[key][i].address + "\n";
                    count++;
                }
            }
        }

        fs.writeFileSync(cnfPath, cnf);
    }

    function makeClientPinAndSslCert(caKeyPath, caCertPath, certsDir, options, conf) {
        options = options || {};
        options.cn = "vcremote-client"; // NOTE: Changing certificates away from referring to vs-mda-remote may require changes to VS, until the CLI can configure certs appropriately
        var clientCertsPath = path.join(certsDir, "client");
        utilHelper.createDirectoryIfNecessary(clientCertsPath);

        // 6 digit random pin (Math.random excludes 1.0)
        var pin = 100000 + Math.floor(Math.random() * 900000);
        var pinDir = path.join(clientCertsPath, "" + pin);
        var pfxPath = path.join(pinDir, "client.pfx");
        var clientKeyPath = path.join(pinDir, "client-key.pem");
        var clientCertPath = path.join(pinDir, "client-cert.pem");

        utilHelper.createDirectoryIfNecessary(pinDir);
        return makeSelfSignedCert(caKeyPath, caCertPath, clientKeyPath, clientCertPath, options, conf).then(function () {
            return makePfx(caCertPath, clientKeyPath, clientCertPath, pfxPath);
        }).then(function () {
            fs.unlinkSync(clientKeyPath);
            fs.unlinkSync(clientCertPath);
            return pin;
        });
    }

    function makePfx(caCertPath, keyPath, certPath, outPfxPath, options) {
        options = options || {};
        var name = options.name || CERT_DEFAULTS.pfx_name;
        return openSslPromise("pkcs12 -export -in " + certPath + " -inkey " + keyPath + " -certfile " + caCertPath + " -out " + outPfxPath + " -name \'" + name + "\' -password pass:");
    }

    // TODO: Make this explain how to configure vs-mda-client via CLI?
    function printVisualStudioSetupToConsole(conf, pin) {
        var host = conf.get("hostname") || os.hostname();
        var port = conf.get("port");
        var pinTimeoutInMinutes = conf.get("pinTimeout");
        console.info(resources.getString("OSXCertSetupInformation", host, port, pin));
        if (pinTimeoutInMinutes) {
            console.info(resources.getString("OSXCertSetupPinTimeout", pinTimeoutInMinutes));
        } else {
            console.info(resources.getString("OSXCertSetupNoPinTimeout"));
        }

        console.info("vcremote generateClientCert");
        console.info("");
    }
    ;
})(Certs || (Certs = {}));

module.exports = Certs;
