﻿/**
* ******************************************************
*                                                       *
*   Copyright (C) Microsoft. All rights reserved.       *
*                                                       *
*******************************************************
*/
/// <reference path="../../typings/node.d.ts" />
/// <reference path="../../typings/Q.d.ts" />
/// <reference path="../../typings/nconf.d.ts" />
/// <reference path="../../typings/vcremote-utils.d.ts" />
/// <reference path="../../typings/express.d.ts" />
/// <reference path="../../typings/express-extensions.d.ts" />
"use strict";
var express = require("express");

var http = require("http");
var https = require("https");

var Q = require("q");

var buildManager = require("./build-manager");
var debugManager = require("./debug-manager");
var HostSpecifics = require("./host-specifics");
var utils = require("vcremote-utils");

var resources = utils.ResourcesManager;

var Server;
(function (Server) {
    var serverInstance;
    var serverConf;

    var errorShutdown;
    var shutdown;

    function start(conf) {
        var app = express();
        app.set("port", conf.get("port"));
        app.set("serverDir", conf.get("serverDir"));
        app.use(express.logger("dev")); // More recent versions of express do not include this functionality, instead it has been moved into separate packages.
        app.use(express.errorHandler());

        // Initialize the build manager (after our app settings are all setup)
        buildManager.init(conf);

        app.get("/", function (req, res) {
            res.send(200, resources.getStringForLanguage(req, "IndexPageContent", app.get("port")));
        });

        // These routes are fully secured through client cert verification:
        if (utils.UtilHelper.argToBool(conf.get("secure"))) {
            app.all("/projects", onlyAuthorizedClientRequest);
            app.all("/projects/*", onlyAuthorizedClientRequest);
            app.all("/debug/*", onlyAuthorizedClientRequest);
            app.all("/filesystem/*", onlyAuthorizedClientRequest);
        }

        app.post("/projects", buildManager.addNewProject);
        app.post("/projects/:id", buildManager.updateProject);
        app.post("/projects/:id/configs", buildManager.addProjectConfig);
        app.post("/projects/:id/build", buildManager.buildProject);
        app.post("/projects/:id/build/:cfg", buildManager.buildProject);
        app.get("/projects", buildManager.getProjects);
        app.get("/projects/devices", buildManager.getDevices);
        app.get("/projects/:id", buildManager.getProjectConfigs);
        app.get("/projects/:id/build/configs", buildManager.getProjectConfigs);
        app.get("/projects/:id/config/:cfg", buildManager.getProjectConfig);
        app.get("/projects/:id/build/:cfg", buildManager.retrieveBuild);
        app.get("/projects/:id/build/:cfg/log", buildManager.getBuildLog);
        app.get("/projects/:id/build/:cfg/install", buildManager.installBuild);
        app.get("/projects/:id/build/:cfg/status", buildManager.getBuildStatus);

        app.get("/filesystem/open", buildManager.execOpen);
        app.post("/filesystem/getfiles", buildManager.getFiles);
        app.post("/filesystem/putfiles", buildManager.putFiles);
        app.post("/filesystem/deletefiles", buildManager.deleteFiles);

        app.get("/debug/setupForDebugging", debugManager.setupForDebugging);
        app.get("/debug/appRemotePath", debugManager.getAppRemotePath);

        app.get("/certs/:pin", downloadClientCerts);

        return startupServer(conf, app).then(registerShutdownHooks).fail(function (err) {
            console.error(resources.getStringForLanguage(conf.get("lang"), "ServerStartFailed"), err);
            if (err.stack) {
                console.error(err.stack);
            }

            throw err;
        });
    }
    Server.start = start;

    function stop(callback) {
        process.removeListener("uncaughtException", errorShutdown);
        process.removeListener("SIGTERM", shutdown);
        process.removeListener("SIGINT", shutdown);
        if (serverInstance) {
            var tempInstance = serverInstance;
            serverInstance = null;
            tempInstance.close(callback);
        }
    }
    Server.stop = stop;

    function resetServerCert(conf) {
        return HostSpecifics.hostSpecifics.resetServerCert(conf);
    }
    Server.resetServerCert = resetServerCert;

    function generateClientCert(conf) {
        return HostSpecifics.hostSpecifics.generateClientCert(conf);
    }
    Server.generateClientCert = generateClientCert;

    function startupServer(conf, app) {
        var isSsl = utils.UtilHelper.argToBool(conf.get("secure"));
        return isSsl ? startupHttpsServer(conf, app) : startupPlainHttpServer(conf, app);
    }

    function startupPlainHttpServer(conf, app) {
        return Q(http.createServer(app)).then(function (svr) {
            var deferred = Q.defer();
            svr.on("error", function (err) {
                deferred.reject(friendlyServerListenError(err, conf));
            });
            svr.listen(app.get("port"), function () {
                serverInstance = svr;
                serverConf = conf;
                console.log(resources.getStringForLanguage(conf.get("lang"), "InsecureServerStarted"), app.get("port"));
                deferred.resolve(svr);
            });
            return deferred.promise;
        });
    }

    function startupHttpsServer(conf, app) {
        var generatedNewCerts = false;
        var generatedClientPin;
        return HostSpecifics.hostSpecifics.initializeServerCerts(conf).then(function (certStore) {
            if (certStore.newCerts) {
                generatedNewCerts = true;
                conf.set("suppressVisualStudioMessage", true);
                return HostSpecifics.hostSpecifics.generateClientCert(conf).then(function (pin) {
                    generatedClientPin = pin;
                    return certStore;
                });
            }

            return Q(certStore);
        }).then(function (certStore) {
            var sslSettings = {
                key: certStore.getKey(),
                cert: certStore.getCert(),
                ca: certStore.getCA(),
                requestCert: true,
                rejectUnauthorized: false
            };
            return https.createServer(sslSettings, app);
        }).then(function (svr) {
            var deferred = Q.defer();
            svr.on("error", function (err) {
                if (generatedNewCerts) {
                    HostSpecifics.hostSpecifics.removeAllCertsSync(conf);
                }
                deferred.reject(friendlyServerListenError(err, conf));
            });
            svr.listen(app.get("port"), function () {
                serverInstance = svr;
                serverConf = conf;
                console.log(resources.getStringForLanguage(conf.get("lang"), "SecureServerStarted"), app.get("port"));
                deferred.resolve(svr);
            });
            return Q(deferred.promise);
        });
    }

    function friendlyServerListenError(err, conf) {
        if (err.code === "EADDRINUSE") {
            return resources.getStringForLanguage(conf.get("lang"), "ServerPortInUse", conf.get("port"));
        } else {
            return err.toString();
        }
    }

    function registerShutdownHooks() {
        // We should have serverConf defined at this point, but just in case fall back to english
        var lang = serverConf && serverConf.get("lang") || "en";

        // It is strongly recommended in a NodeJs server to kill the process off on uncaughtException.
        errorShutdown = function (err) {
            console.error(resources.getStringForLanguage(lang, "UncaughtErrorShutdown"));
            console.error(err);
            console.error(err.stack);
            console.info(resources.getStringForLanguage(lang, "ServerShutdown"));
            buildManager.shutdown();
            debugManager.shutdown();
            process.exit(1);
        };
        process.on("uncaughtException", errorShutdown);

        // Opportunity to clean up builds on exit
        shutdown = function () {
            console.info(resources.getStringForLanguage(lang, "ServerShutdown"));

            // BUG: Currently if buildManager.shutdown() is called while a build log is being written, rimraf will throw an exception on windows
            buildManager.shutdown();
            debugManager.shutdown();
            serverInstance.close();
            process.exit(0);
        };
        process.on("SIGTERM", shutdown);
        process.on("SIGINT", shutdown);
    }

    function onlyAuthorizedClientRequest(req, res, next) {
        if (!req.client.authorized) {
            res.send(401, resources.getStringForLanguage(req, "UnauthorizedClientRequest"));
        } else {
            next();
        }
    }
    ;

    // Downloads client SSL certs (pfx format) for pin specified in request
    function downloadClientCerts(req, res) {
        HostSpecifics.hostSpecifics.downloadClientCerts(req, res);
    }
})(Server || (Server = {}));

module.exports = Server;
